# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-19 03:47
from hanlp.common.constant import HANLP_URL

FLAIR_LM_FW_WMT11_EN = HANLP_URL + 'lm/flair_lm_wmt11_en_20200211_091932.zip#flair_lm_fw_wmt11_en'
FLAIR_LM_BW_WMT11_EN = HANLP_URL + 'lm/flair_lm_wmt11_en_20200211_091932.zip#flair_lm_bw_wmt11_en'

ALL = {}