# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-28 19:10
from hanlp.pretrained import cws
from hanlp.pretrained import dep
from hanlp.pretrained import sdp
from hanlp.pretrained import glove
from hanlp.pretrained import pos
from hanlp.pretrained import rnnlm
from hanlp.pretrained import word2vec
from hanlp.pretrained import ner
from hanlp.pretrained import classifiers

# Will be filled up during runtime
ALL = {}
