# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-28 22:51

CTB5_POS_HOME = 'http://file.hankcs.com/corpus/ctb5.1-pos.zip'

CTB5_POS_TRAIN = f'{CTB5_POS_HOME}#train.tsv'
CTB5_POS_VALID = f'{CTB5_POS_HOME}#dev.tsv'
CTB5_POS_TEST = f'{CTB5_POS_HOME}#test.tsv'
