# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-28 00:51

SEMEVAL2016 = 'https://github.com/HIT-SCIR/SemEval-2016/archive/master.zip'

SEMEVAL2016_NEWS_TRAIN = SEMEVAL2016 + '#train/news.train.conll'
SEMEVAL2016_NEWS_VALID = SEMEVAL2016 + '#validation/news.valid.conll'
SEMEVAL2016_NEWS_TEST = SEMEVAL2016 + '#test/news.test.conll'

SEMEVAL2016_TEXT_TRAIN = SEMEVAL2016 + '#train/text.train.conll'
SEMEVAL2016_TEXT_VALID = SEMEVAL2016 + '#validation/text.valid.conll'
SEMEVAL2016_TEXT_TEST = SEMEVAL2016 + '#test/text.test.conll'
